@@ -33,12 +33,12 @@
 #include "vino-input.h"
 #include "vino-cursor.h"
 #include "vino-prompt.h"
-#include "vino-status-icon.h"
 #include "vino-dbus-listener.h"
 #include "vino-util.h"
 #include "vino-enums.h"
 #include <sys/poll.h>
 #include <dbus/dbus-glib.h>
+#include <gtk/gtk.h>
 
 #ifdef VINO_ENABLE_KEYRING
 #include <gnome-keyring.h>
@@ -171,9 +171,6 @@ static void
 vino_server_client_accepted (VinoServer *server,
                              VinoClient *client)
 {
-  if (!server->priv->icon)
-    server->priv->icon = vino_status_icon_new (server, server->priv->screen);
-
   vino_status_icon_add_client (server->priv->icon, client);
 }
 
@@ -181,14 +178,10 @@ static void
 vino_server_client_disconnected (VinoServer *server,
                                  VinoClient *client)
 {
-  if (vino_status_icon_remove_client (server->priv->icon, client))
-    {
-      g_object_unref (server->priv->icon);
-      server->priv->icon = NULL;
-
-      vino_server_lock_screen(server);
-    }
+  vino_status_icon_remove_client (server->priv->icon, client);
+  vino_server_lock_screen (server);
 }
+
 static void
 vino_server_handle_client_gone (rfbClientPtr rfb_client)
 {
@@ -774,12 +767,41 @@ vino_server_release_framebuffer (VinoServer *server)
   server->priv->fb = NULL;
 }
 
+static void
+vino_server_clipboard_cb (GtkClipboard *cb,
+                          GdkEvent     *event,
+                          VinoServer   *server)
+{
+  gchar *text, *out;
+  size_t a, b;
+
+  if (server->priv->on_hold)
+    return;
+
+  text = gtk_clipboard_wait_for_text (cb);
+  if (!text)
+    return;
+
+  out = g_convert (text, -1, "iso8859-1", "utf-8", &a, &b, NULL);
+  if (out)
+    {
+      rfbSendServerCutText (server->priv->rfb_screen,
+                            out,
+                            strlen (out));
+      g_free (out);
+    }
+
+  g_free (text);
+
+}
+
 static void
 vino_server_init_from_screen (VinoServer *server,
 			      GdkScreen  *screen)
 {
   rfbScreenInfoPtr  rfb_screen;
   char             *name;
+  GtkClipboard     *cb;
 
   g_return_if_fail (server->priv->screen == NULL);
   g_return_if_fail (screen != NULL);
@@ -871,6 +893,15 @@ vino_server_init_from_screen (VinoServer *server,
   vino_mdns_add_service ("_rfb._tcp", rfb_screen->rfbPort);
 
   server->priv->listener = vino_dbus_listener_new (server);
+
+  cb = gtk_clipboard_get_for_display (gdk_screen_get_display (screen),
+                                      GDK_SELECTION_CLIPBOARD);
+  g_signal_connect (cb,
+                    "owner-change",
+                    G_CALLBACK (vino_server_clipboard_cb),
+                    server);
+
+  server->priv->icon = vino_status_icon_new (server, server->priv->screen);
 }
 
 static void
@@ -924,6 +955,10 @@ vino_server_finalize (GObject *object)
   if (server->priv->listener)
     g_object_unref (server->priv->listener);
   server->priv->listener = NULL;
+
+  if (server->priv->icon)
+    g_object_unref (server->priv->icon);
+  server->priv->icon = NULL;
   
   g_free (server->priv);
   server->priv = NULL;
@@ -1401,6 +1436,9 @@ vino_server_set_on_hold (VinoServer *server,
 	}
 
       g_object_notify (G_OBJECT (server), "on-hold");
+      
+      if (server->priv->icon)
+        vino_status_icon_update_state (server->priv->icon);
     }
 }
 
@@ -1576,3 +1614,11 @@ vino_server_set_lock_screen (VinoServer *server,
       g_object_notify (G_OBJECT (server), "lock-screen");
     }
 }
+
+VinoStatusIcon *
+vino_server_get_status_icon (VinoServer *server)
+{
+  g_return_val_if_fail (VINO_IS_SERVER (server), NULL);
+
+  return server->priv->icon;
+}
